% ---------------------------------------------------------------------
%   COPYRIGHT 2001-2024 MICHAEL TSIROULNIKOV. ALL RIGHTS RESERVED 
%   DIRECT OR INDIRECT COMMERCIAL DERIVATIVES OR USAGE PROHIBITED
%   GNU General Public License v.3+ <https://www.gnu.org/licenses/>
% ---------------------------------------------------------------------
classdef fsaf
    methods(Static)
        %% ---------------------------------------------------------------------
        function [in,mask]=gen(sz,varargin)
            if length(sz) == 1
                nse=0;
                n=1;
                mask=ones(sz,1);
                while n <= nargin-1
                    imq=varargin{n};
                    param=varargin{n+1};
                    if strcmpi(imq, 'on')
                        szon=param; % 
                    elseif strcmpi(imq,'off')
                        szoff=param;
                    elseif strcmpi(imq,'nse')
                        nse=param;
                    elseif strcmpi(imq,'lpf')
                        lpf=param;
                    elseif strcmpi(imq,'norm')
                        norm=param;
                    elseif strcmpi(imq,'mask')
                        mask=param;
                    end
                    n=n+2;
                end
                if exist('szon','var') 
                    if exist('szoff','var')
                        szr=szon+szoff;
                        rep=floor(sz/szr);
                        for r=1:rep
                            st=(r-1)*szr;
                            mask(st+1:st+szon)=1;
                            mask(st+szon+1:st+szr)=nse;
                        end
                    else
                        if sz > szon
                            mask(szon+1:end)=nse;
                        end
                    end
                    if exist('lpf','var')
                        [b,a]=butter(1,lpf);
                        mask=filter(b,a,mask);
                    end
                end
                if exist('norm','var') 
                    mask=mask*norm;
                end
            else
                mask=sz;
                sz=length(sz);
            end
            in=randn(sz,1).*mask;
        end
        %% ---------------------------------------------------------------------
        function [rir,w]=rir_make(RT60,LEN, DLY,HPF,LPF)
            % RT60 must be in samples
            if nargin < 5
                LPF=0.9;
            end
            if nargin < 4
                HPF=0.003;
            end
            if nargin < 3
                DLY=0;
            end
            if nargin < 2
                LEN=RT60;
            end
            LEN=round(LEN);
            DLY=round(DLY);
            % RT60 in samples
            rir=randn(LEN,1);
            w=ones(LEN,1);
            coef=1e-3^(1/RT60);
            for k=2:LEN
                w(k)=w(k-1)*coef;
            end
            if (DLY > 0)
                w(1:DLY)=eps; % to avoid NAN
            end
            rir=rir.*w;
            if (DLY > 0)
                rir(1:DLY)=0;
            end
            if HPF > 0
                [b,a]=butter(2,HPF,'high');
                rir=filter(b,a,rir);
            end
            if LPF > 0
                [b,a]=butter(2,LPF);
                rir=filter(b,a,rir);
            end
            rir=rir/norm(rir);
            w=w/sum(w);
        end
        %% ---------------------------------------------------------------------
        function [rir,P0]=rir_jtf(RT60,LEN,b,a,LADF)
            if nargin < 5
                LADF=LEN;
            end
            if nargin < 4
                a=1;
            end
            if nargin < 3
                b=1;
            end
            if length(RT60) == 1
                % RT60 must be in samples
                gamma=1e-3^(1/RT60);
                RT60=ones(LEN,1);
                for k=2:LEN
                    RT60(k)=RT60(k-1)*gamma; 
                end
            end
            LEN=min(LEN,length(RT60));
                %D=diag(sqrt(RT60));
%                irir=1; % w already been scaled
            irir=1/norm(RT60);
            rir=randn(LEN,1);
            rir=RT60.*rir;
            rir=filter(b,a,rir);
            rir=rir*irir;
            % ---------------------------------------------------------------------
            LADF=min(LADF,LEN);
            RT60=RT60(1:LADF);
            %D=D(1:LADF,1:LADF);
            P0=diag(RT60.*RT60); % squared
            %P0=D'*P0*D;
            if length(b)==1 && length(a)==1 && b~=1 && a~=1
                imp=impz(b,a,LADF);
                F=zeros(LADF); 
                for n=1:LADF
                    F(n,n:end)=imp(1:LADF+1-n); 
                end
                P0=F'*P0*F;
            end
            P0=P0*irir^2;
        end
        %% -------------------------------------------------------------------
        function fmin = minphase_make(fir,Nroots,ButOrder,naa)
            if nargin < 5
                naa=40;
            end
% you have to understand what you are doing            
            if nargin < 4
                ButOrder=1;
            end
            if nargin < 3
                Nroots=64;
            end
            z=size(fir); if z(1) < z(2), fir=fir';end
            lenin=length(fir);
            if lenin > Nroots
                MR=ceil(lenin/Nroots);
                if ButOrder > 0
                    [b,a]=butter(ButOrder,1/MR);
                end
                f=resample(fir,1,MR,naa);
                f=polystab(f)';
                f=resample(f,MR,1,naa);
                if ButOrder > 0
                    f=filter(b,a,f);
                end
            else
                f=polystab(fir)';
                while length(f) < lenin
                    f=[f; 0];
                end
            end
            f=f*sign(sum(f));
            fmin=f/sum(f); % 0dB DC
            sz=size(fmin); if sz(1) < sz(2), fmin=fmin';end
            fmin=flipud(fmin);
        end
        %% ---------------------------------------------------------------------
        function [IS,s] = invS(S,thr,p)
            dS=real(diag(S)); % not-negative definite 
            maxS=max(dS);
            s=dS/maxS;
            LL=length(s);
            IS=eye(LL);
            if thr > 0
                thrP=thr^p;
            elseif thr < 0
                thrP=10^(thr*p/20);
            else
                thrP=0;
            end
            for n=1:LL
                sn=s(n);
                snP1=sn^(p-1);
                isn = snP1 / (thrP+sn*snP1);
                IS(n,n)=isn/maxS;
            end
        end
        %% ---------------------------------------------------------------------
        function [IA,s,U,S,V] = inv(A,thr,p)
            if nargin < 3
                p=2;
            else
                p=max(p,1); % norm from 1 -> +inf
            end
            if nargin < 2
                thr=0;
            else
                thr=max(thr,0);
            end
            [U,S,V]=svd(A,'econ');
            [IS,s]=fsaf.invS(S,thr,p);
            IA=V*IS*U';
        end
        %% ---------------------------------------------------------------------
        function [x,s] = linresolve(A,b,InvD,thr,p)
            if nargin < 5
                p=2;
            else
                p=max(p,1); % norm from 1 -> +inf
            end
            if nargin < 4
                thr=0;
            else
                thr=max(thr,0);
            end
            if nargin > 2
                AxA=A'*A;
                if size(InvD,2) == 1
                    if size(InvD,1) == 1
                        LL=max(size(AxA));
                        InvD=eye(LL)*invD;
                    else
                        InvD=diag(InvD);
                    end
                end
                AxA=AxA+InvD;
                thr=thr^2;
                [IA,s]=fsaf.inv(AxA,thr,p);
                Ab=A'*b;
                x=IA*Ab;
            else
                [IA,s]=fsaf.inv(A,thr,p);
                x=IA*b;
            end
        end
        %% ---------------------------------------------------------------------
        function [x,s] = linsolve(A,b,thr,p)
            if nargin < 4
                p=2;
            else
                p=max(p,1); % norm from 1 -> +inf
            end
            if nargin < 3
                thr=0;
            else
                thr=max(thr,0);
            end
            
            [IA,s]=fsaf.inv(A,thr,p);
            x=IA*b;
        end
        %% ---------------------------------------------------------------------
        function f=fig(n)
            if n > 0
                f = figure(n);
            else
                f = figure;
            end
            fp=f.Position;
            if exist('fsaf_cfg','var')
                fp(3)=fsaf_cfg.szx;
                fp(4)=fsaf_cfg.szy;
            else
                fp(3)=672;
                fp(4)=384;
            end
            f.Position=fp;
        end
        %% ---------------------------------------------------------------------
        function limy = db_limits()
            if exist('fsaf_cfg','var')
                limy=[fsaf_cfg.miny fsaf_cfg.maxy];
            else
                limy = [-92 2];
            end
        end
        %% ---------------------------------------------------------------------
        function limdb(limy)
            if nargin < 1
                limy = fsaf.db_limits();
            end
            ylim(limy);
            grid on;
            ylabel('level, dB');
        end
        %% ---------------------------------------------------------------------
        function dblim(limy)
            if nargin < 1
                fsaf.limdb();
            else
                fsaf.limdb(limy);
            end
            xlabel('frequency, f/f_{N}');
        end
        %% ---------------------------------------------------------------------
        function f=tddb_plot(figno,imq,varargin)
            f=fsaf.td_plot(-figno,imq,varargin{:});
        end
        %% ---------------------------------------------------------------------
        function hx=clr(idx)
             a={'#0072BD','#D95319','#EDB120',...
                '#7E2F8E','#77AC30','#4DBEEE',...
                '#00365D','#792309','#000000'};%,'#7D6110'};
            idx=1+rem(idx-1,length(a));
            hx=a{idx};
        end
        %% ---------------------------------------------------------------------
        function hx=clrx(idx)
            if nargin < 1 || idx < 1
                idx=10;
            end
            %'#D95319'
            a={'#D90000','#EDB120','#77AC30',...
               '#4DBEEE','#0072BD','#7E2F8E',...
               '#000055','#005500','#770000','#000000'};
            idx=1+rem(idx-1,length(a));
            hx=a{idx};
        end
        %% ---------------------------------------------------------------------
        function f=td_plot(figno,imq,varargin)
            NADD=(nargin-2)/2;
            lenin=1;
            maxin=0;
            alignx=0;
            lines=0;
            hlbrt=0;
            for n=1:NADD
                imqn=varargin{2*n-1};
                x=varargin{2*n-0};
                lenx=length(x);
                if strcmpi(imqn,'tnorm') 
                    tnorm=x;
                elseif strcmpi(imqn,'tshift')
                    tshift=x;
                elseif strcmpi(imqn,'ynorm')
                    ynorm=x;
                elseif strcmpi(imqn,'miny') || strcmpi(imqn,'ymin')
                    miny=x;
                elseif strcmpi(imqn,'maxy') || strcmpi(imqn,'ymax')
                    maxy=x;
                elseif strcmpi(imqn,'xalign') || strcmpi(imqn,'alignx')
                    alignx=x;
                elseif strcmpi(imqn,'title')
                    ttl=x;
                elseif strcmpi(imqn,'logx') || strcmpi(imqn,'xlog')
                    xlog=x;
                elseif strcmpi(imqn,'m')
                    M=x;
                elseif strcmpi(imqn,'hilbert')
                    hlbrt=x;
                    if x
                        imq=sprintf('^H %s',imq);
                    end
                else
                    lines=lines+1;
                    entries{lines}=imqn;
                    X{lines}=x;
                    lenin=max(lenin,lenx);
                    maxin=max(maxin,max(abs(x)));
                end
            end
            
            t=(1:lenin);
            if exist('M','var')
                if alignx == 0
                    t=(-lenin/2+0.5:lenin/2-0.5)/M;
                else
                    t=(1:lenin)/M;
                end
            else
                if exist('tnorm','var')
                    t=t/tnorm;
                else
                    t=t/lenin;
                end
                if exist('tshift','var')
                    t=t-tshift;
                end
            end
            f=fsaf.fig(abs(figno));
            for n=1:lines
                x=X{n};
                lenx=length(x);
                lt='-';
                if n > 10
                    lt='--';
                end
                if lenx > 1
                    if alignx == 0
                        shl=(lenin-lenx)/2;
                        shr=shl;
                        if ~isinteger(shl)
                            shr=ceil(shl);
                            shl=floor(shl);
                        end
                    elseif alignx < 0
                        shl=0;
                        shr=(lenin-lenx)/1;
                    else
                        shl=(lenin-lenx)/1;
                        shr=0;
                    end
                    if exist('M','var')
                        x=x*M;
                    end
                    if figno > 0
                        if hlbrt
                            lenx2=round(lenx/8);
                            xa=vertcat(zeros(lenx2,1),x,zeros(lenx2,1));
                            xa=hilbert(xa);
                            x=xa(lenx2+1:end-lenx2);
                            x=imag(x);
                        end
                        if exist('ynorm','var') 
                            if ynorm > 0
                                x=x/max(abs(x));
                            elseif ynorm < 0
                                x=x/maxin;
                            end
                        end
                    else % < 0, ie tddb
                        if hlbrt
                            lenx2=round(lenx/8);
                            xa=vertcat(zeros(lenx2,1),x,zeros(lenx2,1));
                            xa=hilbert(xa);
                            x=xa(lenx2+1:end-lenx2);
                        end
                        x=db(x);
                        if exist('ynorm','var') 
                            if ynorm > 0
                                x=x-max(x);
                            elseif ynorm < 0
                                x=x-db(maxin);
                            end
                        end
                    end
                    if exist('xlog','var')
                        semilogx(t(1+shl:end-shr),x,lt,'Color',fsaf.clr(n));
                    else
                        plot(t(1+shl:end-shr),x,lt,'Color',fsaf.clr(n));
                    end
                end
                hold on;
            end
            grid on;
            hold off;
            legend(entries);
            legend('boxoff');
            if exist('M','var')
                xlabel('time/M');
                if figno > 0
                    ylabel('value*M');
                else
                    ylabel('level, dB');
                end
            else
                xlabel('time');
                if exist('ynorm','var') && ynorm > 0
                    ylabel('value, normalized');
                else
                    if figno > 0
                        ylabel('value');
                    else
                        ylabel('level, dB');
                    end
                end
            end
            if figno < 0
                ylim([-102 2]);
            end
            if exist('miny','var') || exist('maxy','var')
                limy = fsaf.db_limits();    
                if exist('miny','var')
                    limy(1)=miny;
                end
                if exist('maxy','var')
                    limy(2)=maxy;
                end
                ylim(limy);
            end
            if ~exist('M','var')
                xlim([0 max(t)]);
            else
                xlim([min(t) max(t)]);
            end
            str=sprintf('%s in time domain',imq);
            if figno < 0
                str=sprintf('%s, dB scale', str);
            end
            if exist('ttl','var')
                str=sprintf('%s %s',ttl,str);
            end
            title(str);
        end
        %% ---------------------------------------------------------------------
        function fg=fd_plot(figno,sTitle,varargin)
            NADD=(nargin-2)/2;
            fmax=1;
            fmin=0;
            iLog=0;
            ymin=-102;
            ymax=2;
            FS=2;
            nfm=100;
            M=1;
            nf=nfm*M;
            lenin=1;
            lines=0;
            normeach=0;
            normnrg=0;
            str='in frequency domain';
            for n=1:NADD
                imq=varargin{2*n-1};
                x=varargin{2*n-0};
                lenx=length(x);
                if strcmpi(imq,'fs') 
                    FS=x;
                elseif strcmpi(imq,'fmax')
                    fmax=x;
                elseif strcmpi(imq,'fmin')
                    fmin=x;
                elseif strcmpi(imq,'normeach')
                    normeach=x;
                elseif strcmpi(imq,'normnrg')
                    normnrg=x;
                elseif strcmpi(imq,'nfm')
                    nfm=x;
                    nf=nfm*M;
                elseif strcmpi(imq,'log')
                    iLog=1;
                    fmin=x;
                elseif strcmpi(imq,'title')
                    sTitle=x;
                elseif strcmpi(imq,'m')
                    M=x;
                    nf=nfm*M;
                    fmax=M;
                else
                    lenin=max(lenin,lenx);
                    entries{n}=imq;
                    lines=lines+1;
                end
            end % end-of-for lines
            if rem(nargin-2,2) 
                imq=varargin{nargin-2};
                if strcmpi(imq,'NEAR')
                    fmax=2;
                end
            end

            if FS == 0
                FS = M*2;
            end
            
            fg=fsaf.fig(figno);
            for n=1:lines
                imq=varargin{2*n-1};
                x=varargin{2*n-0};
                lenx=length(x);
                if strcmpi(imq,'1/eq')
                    f=(1:lenx)'/lenx;
                    dh=x;
                else
                    [h,f]=freqz(x,1,nf,FS*M);
                    dh=db(h);
                end
                if normeach
                    dh=dh-max(dh);
                end
                if normnrg
                    dh=dh/2;
                end
                if iLog
                    semilogx(f,dh);
                else
                    plot(f,dh);
                end
                if n == 1 && lines > 1
                    hold on;
                end
            end % end-of-for lines
            grid on;
            if lines > 1
                hold off;
            end
            xlim([fmin fmax]);
            fsaf.dblim([ymin ymax]);
            
            legend(entries,'location','best');
            legend('boxoff');
            if exist('M','var')
                xlabel('frequency, f/f_{N}');
            else
                xlabel('frequency');
            end
            if normeach
                ylabel('normalized level, dB');
            else
                ylabel('level, dB');
            end
            if ~isempty(sTitle)
                str=sprintf('%s %s',sTitle,str);
            end
            title(str);
        end % end of function
    end % end of methods
end % end of class
        
